# originally inspired by "six" by Benjamin Peterson

import inspect
import sys


if sys.version_info < (3,0):
    text_type = unicode
    binary_type = str

    import StringIO
    StringIO = BytesIO = StringIO.StringIO
else:
    text_type = str
    binary_type = bytes

    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO


def getargspec_permissive(func):
    """
    An `inspect.getargspec` with a relaxed sanity check to support Cython.

    Motivation:

        A Cython-compiled function is *not* an instance of Python's
        types.FunctionType.  That is the sanity check the standard Py2
        library uses in `inspect.getargspec()`.  So, an exception is raised
        when calling `argh.dispatch_command(cythonCompiledFunc)`.  However,
        the CyFunctions do have perfectly usable `.func_code` and
        `.func_defaults` which is all `inspect.getargspec` needs.

        This function just copies `inspect.getargspec()` from the standard
        library but relaxes the test to a more duck-typing one of having
        both `.func_code` and `.func_defaults` attributes.
    """
    if inspect.ismethod(func):
        func = func.im_func

    # Py2 Stdlib uses isfunction(func) which is too strict for Cython-compiled
    # functions though such have perfectly usable func_code, func_defaults.
    if not (hasattr(func, "func_code") and hasattr(func, "func_defaults")):
        raise TypeError('{!r} missing func_code or func_defaults'.format(func))

    args, varargs, varkw = inspect.getargs(func.func_code)
    return inspect.ArgSpec(args, varargs, varkw, func.func_defaults)


if sys.version_info < (3,0):
    getargspec = getargspec_permissive
else:
    # in Python 3 the basic getargspec doesn't support keyword-only arguments
    # and annotations and raises ValueError if they are discovered
    getargspec = inspect.getfullargspec



# coding: utf-8
"""
Unit Tests For Utility Functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import functools
import sys

import pytest


from argh.utils import get_arg_spec


def function(x, y=0):
    return

def decorated(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        print("Wrapping function call")
        return f(*args, **kwargs)
    return wrapped


def _assert_spec(f, **overrides):
    spec = get_arg_spec(f)

    defaults = {
        'args': ['x', 'y'],
        'varargs': None,
        'varkw': None,
        'defaults': (0,),
        'kwonlyargs': [],
        'annotations': {},
    }

    for k in defaults:
        actual = getattr(spec, k)
        expected = overrides[k] if k in overrides else defaults[k]
        assert actual == expected


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__plain_func():
    _assert_spec(function)


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__decorated_func():
    def d(_f):
        return _f
    decorated = d(function)

    _assert_spec(decorated)


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__wrapped():
    wrapped = decorated(function)
    _assert_spec(wrapped)


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__wrapped_nested():
    wrapped = decorated(decorated(function))
    _assert_spec(wrapped)


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__wrapped_complex():
    def wrapper_deco(outer_arg):
        def _outer(f):
            @functools.wraps(f)
            def _inner(*args, **kwargs):
                return f(*args, **kwargs)
            return _inner
        return _outer

    wrapped = wrapper_deco(5)(function)

    _assert_spec(wrapped)


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__static_method():
    class C:
        @staticmethod
        def f(x, y=0):
            return x

    _assert_spec(C.f)


@pytest.mark.skipif(sys.version_info < (3,5),
                    reason="requires python3.5")
def test_get_arg_spec__method():
    class C:
        def f(self, x, y=0):
            return x

    _assert_spec(C.f, args=['self', 'x', 'y'])



# coding: utf-8
"""
Common stuff for tests
~~~~~~~~~~~~~~~~~~~~~~
"""
import os
import sys
from collections import namedtuple

from argh import ArghParser
from argh.compat import BytesIO, StringIO


_CmdResult = namedtuple('CmdResult', ('out', 'err', 'exit'))
# hacky constructor for default exit value
def CmdResult(out, err, exit=None):
    return _CmdResult(out, err, exit)


class DebugArghParser(ArghParser):
    "(does not print stuff to stderr on exit)"

    def exit(self, status=0, message=None):
        raise SystemExit(message)

    def error(self, message):
        self.exit(2, message)


def make_IO():
    "Returns a file object of the same type as `sys.stdout`."
    if sys.version_info < (3,0):
        return BytesIO()
    else:
        return StringIO()


def call_cmd(parser, command_string, **kwargs):
    if hasattr(command_string, 'split'):
        args = command_string.split()
    else:
        args = command_string

    io_out = make_IO()
    io_err = make_IO()

    if 'output_file' not in kwargs:
        kwargs['output_file'] = io_out
    kwargs['errors_file'] = io_err

    try:
        result = parser.dispatch(args, **kwargs)
    except SystemExit as e:
        result = None
        exit = e.code or 0 # e.code may be None
    else:
        exit = None

    if kwargs.get('output_file') is None:
        return CmdResult(out=result, err=io_err.read(), exit=exit)
    else:
        io_out.seek(0)
        io_err.seek(0)
        return CmdResult(out=io_out.read(), err=io_err.read(), exit=exit)


def run(parser, command_string, kwargs=None, exit=False):
    """ Calls the command and returns CmdResult(out, err, exit status)
    with exit status None if no SystemExit was raised.

    :exit:
        if set to `True` and a SystemExit is raised, the status code is returned;
        if the exception is not raised, an AssertionError is raised.

    """
    kwargs = kwargs or {}
    result = call_cmd(parser, command_string, **kwargs)
    if exit:
        if result.exit is None:
            raise AssertionError('Did not exit')
        else:
            return result.exit
    return result


def get_usage_string(definitions='{cmd} ...'):
    prog = os.path.basename(sys.argv[0])
    return 'usage: ' + prog + ' [-h] ' + definitions + '\n\n'
